from typing import Any, List, Dict, Tuple, Callable

class LengthValidator():
    def __init__(self, min_length = 6, max_length = 100, error_message: str = '') -> None:
        self.min_length = min_length
        self.max_length = max_length
        self.error_message = error_message

    def __call__(self, key_name: str, data: str) -> str:
        if len(data) < self.min_length or len(data) > self.max_length:
            return self.error_message or f'{key_name}长度不符合'


class ValidatorField():
    '''
    校验字段
    ValidatorField(
        required: '用户名必填',
        ftype: str,
        custom_validators: [
            LengthValidator(error_message='用户名长度在6~12个字符！')
        ]
    )
    '''
    def __init__(self, required: str = '', ftype: Any = None, custom_validators: List[Callable] = []) -> None:
        self.required = required
        self.ftype = ftype
        self.custom_validators = custom_validators

    def __call__(self, key_name: str, data: Dict) -> List[str]:
        errs = []
        exist = key_name in data

        # 必填校验
        if self.required and not exist:
            errs.append(self.required)

        # 自定义校验
        if exist:
            for v in self.custom_validators:
                errmsg = v(key_name, data[key_name])
                errmsg and errs.append(errmsg)

        return errs


class Validator():
    '''
    校验器
    Validator({
        user_name: ValidatorField()
    })
    '''
    def __init__(self, fields: Dict[str, ValidatorField]) -> None:
        self.fields = fields

    def valide(self, data: Dict) -> List[str]:
        if type(data) != dict:
            return ['类型错误']

        errs = []

        for v in self.fields:
            if v in data:
                vf = self.fields[v]
                errs.extend(vf(v, data))

        return errs



'''
Author: liubei
Date: 2021-04-08 17:21:53
LastEditTime: 2021-04-10 11:28:21
Description: 
'''
import json
import hmac
import hashlib
from base64 import urlsafe_b64encode, urlsafe_b64decode

jwt_encode = 'utf-8'
# TODO：可以随机生成secret, secret和jwt的映射关系保存在数据库，解码时获取
secret_str = 'basic_jwt_secret_001'

def create_jwt(payload: dict, secret: str = secret_str) -> str:
    header = json.dumps({
        'alg': 'HS256'
    }).encode(jwt_encode)
    payload = payload if payload else {}
    payload = json.dumps(payload).encode(jwt_encode)

    # base64编码
    header = urlsafe_b64encode(header).decode(jwt_encode)
    payload = urlsafe_b64encode(payload).decode(jwt_encode)

    # 加密
    signature = header + '.' + payload
    signature = hmac.new(
        secret.encode(jwt_encode), signature.encode(jwt_encode), hashlib.sha256).digest()
    signature = urlsafe_b64encode(signature).decode(jwt_encode)

    return header + '.' + payload + '.' + signature


def parse_jwt(jwt_str: str, secret: str = secret_str) -> dict:
    jwt_token = jwt_str.split('.')

    if len(jwt_token) != 3:
        return

    header = jwt_token[0]
    payload = jwt_token[1]
    signature = header + '.' + payload
    signature = hmac.new(
        secret.encode(jwt_encode), signature.encode(jwt_encode), hashlib.sha256).digest()
    signature = urlsafe_b64encode(signature).decode(jwt_encode)

    # jwt被篡改过
    if signature != jwt_token[2]:
        return

    # base64解码
    header = urlsafe_b64decode(header.encode(jwt_encode)).decode(jwt_encode)
    payload = urlsafe_b64decode(payload.encode(jwt_encode)).decode(jwt_encode)

    # 还原出信息
    # header = json.loads(header)
    # payload = json.loads(payload)

    return json.loads(payload)



import asyncio, os, json, uuid
from tornado.web import RequestHandler

class HandleUploadFile(RequestHandler):
    '''
    将上传的文件保存到指定的目录
    '''
    upload_dir = './' # 子类需指定该属性
    origin_file_name = ''
    file_name = ''
    upload_file_error_msg = ''
    file_type = ''
    file_size = 0

    async def prepare(self):
        fut = super().prepare()

        if asyncio.coroutines.iscoroutine(fut):
            await fut

        f = self.request.files['file'] if 'file' in self.request.files else None

        if not f or len(f) < 1:
            self.upload_file_error_msg = '缺少上传文件'
            return

        dirpath = self.upload_dir

        if not os.path.exists(dirpath):
            os.makedirs(dirpath)

        self.origin_file_name = f[0].get('filename')
        self.file_type = os.path.splitext(f[0].get('filename'))[-1]
        self.file_name = str(uuid.uuid1()) + self.file_type
        self.file_size = len(f[0].get('body'))

        with open(dirpath + self.file_name, mode='wb') as sf:
            try:
                sf.write(f[0].get('body'))
            except Exception as e:
                self.upload_file_error_msg = '写入文件失败'
                raise e



from typing import List, Any
import aiomysql
from pymysql import converters
from pymysql.constants import FIELD_TYPE
from ..base.log import Log


charset='utf8'
conv = {
    **converters.conversions,
    # 使用pymysql提供的escape函数进行时间字段的转换
    **{
    FIELD_TYPE.TIMESTAMP: lambda x: converters.escape_item(converters.convert_mysql_timestamp(x), charset).replace("'", ''),
    FIELD_TYPE.DATETIME: lambda x: converters.escape_item(converters.convert_datetime(x), charset).replace("'", ''),
    FIELD_TYPE.TIME: lambda x: converters.escape_item(converters.convert_timedelta(x), charset).replace("'", ''),
    FIELD_TYPE.DATE: lambda x: converters.escape_item(converters.convert_date(x), charset).replace("'",''),
    FIELD_TYPE.DECIMAL: float,
    FIELD_TYPE.NEWDECIMAL: float,
    }
}

dlog = Log()

class MDao():
    def __init__(self, config:dict):
        self._conn = None
        self._config = config

    async def _ensure_connect(self):
        if self._conn is None:
            self._conn = await aiomysql.connect(conv=conv, **self._config)

    async def query(self, sql:str, params=None) -> List[dict]:
        dlog.print(sql, params)
        res = []
        await self._ensure_connect()
        async with self._conn.cursor(aiomysql.DictCursor) as c:
            await c.execute(sql, args=params)
            res = await c.fetchall()
            await self._conn.commit()

        return res

    async def dml(self, sql:str, params=None, many=False):
        dlog.print(sql, params)
        res = None
        await self._ensure_connect()
        async with self._conn.cursor() as c:
            try:
                if not many:
                    res = await c.execute(sql, args=params)
                else:
                    res = await c.executemany(sql, args=params)

                await self._conn.commit()
            except:
                await self._conn.rollback()
                raise

        dlog.print(res)
        return res

    def __del__(self):
        if self._conn:
            self._conn.close()



from enum import Enum
from typing import List, Dict, Optional, Tuple, Union, Any, NewType

from .m_dao import MDao

# 数据库类型定义
VARCHAR = NewType('VARCHAR', str)
DATETIME = NewType('DATETIME', str)
INT = NewType('INT', int)
TINYINT = NewType('TINYINT', int)

# 字段查询时方式
class MPredicate(Enum):
    EQUAL = 1
    IN = 2
    LIKE = 3
    REGEXP = 4
    BETWEEN = 5

class MExpression(Enum):
    OR = 1
    AND = 2

class MBaseT():
    ...


class MBaseSelectExprT():
    '''
    TODO: 查询sql中字段(select_expr)部分
    '''
    def get_select_expr(self):
        pass


class MBaseWhereT():
    '''
    查询sql中的where部分
    '''
    field_names = []
    query_fields = {}

    def get_where_params(self, mt):
        '''
        mt: { filedName: value }
        return: [
            [filedName, Predicate, value, Expression],
            ['user_name', LIKE, 'abc', 'or'],
            ['id', =, 'abc', 'and'],
        ]
        '''
        fields = self.query_fields
        params = []

        for field_name in self.field_names:
            if hasattr(mt, field_name) and field_name in fields:
                v = getattr(mt, field_name)
                predicate = fields[field_name]
                exp = MExpression.OR

                if type(predicate) == tuple:
                    exp = predicate[1]
                    predicate = predicate[0]

                params.append((field_name, predicate, v, exp))

        return params

    def parse_where_sql(self, params):
        '''
        params: [
            ['id', =, '123', MExpression.AND],
            ['user_name', LIKE, 'abc', MExpression.OR],
            ['keyword', LIKE, '111', MExpression.OR],
        ]

        return: (id=123) and (user_name like 'abc' or keyword like '111')
        '''
        andlist = []
        orlist = []

        for field_name, predicate, v, exp in params:
            explist = andlist if exp == MExpression.AND else orlist

            if predicate == MPredicate.LIKE:
                explist.append(f'{field_name} like "%{str(v)}%"')

            elif predicate == MPredicate.EQUAL:
                explist.append(f'{field_name} = \'{str(v)}\'')

            elif predicate == MPredicate.IN:
                if type(v) != list:
                    v = [v]

                v = ', '.join([f'\'{str(vitm)}\'' for vitm in v])
                explist.append(f'{field_name} in ({v})')

            elif v == None:
                explist.append(f'{field_name} is null')

        return ' and '.join([
            f"({' and '.join(andlist) if len(andlist) > 0 else '1=1'})", 
            f"({' or '.join(orlist) if len(orlist) > 0 else '1=1'})"]
        )

class MBaseOrderByT():
    '''
    查询sql中的order by部分
    '''
    field_names = []

    def get_order_by_str(self, mt: MBaseT):
        order_params = []
        order_by = getattr(mt, '_order_by', {})

        for field_name in self.field_names:
            order_type = getattr(order_by, field_name, None)

            if order_type in ['asc', 'desc']:
                order_params.append(f'{field_name} {order_type}')

        return ','.join(order_params)


def normallize_insert_value(inst, k, v):
    assert isinstance(inst, MBaseT)

    ann = inst.__annotations__
    typ = ann.get(k)

    if typ == VARCHAR:
        return f'\'{v}\''
    elif typ == DATETIME:
        return 'null' if not v else f'\'{v}\''
    elif typ in [INT, TINYINT]:
        return str(v)
    elif v == None:
        return 'null'

    return f'\'{str(v)}\''


class MBase(MBaseWhereT, MBaseOrderByT):
    db_config = None
    table_name = None
    query_fields = {}
    field_names = []

    def __init__(self):
        self._db = MDao(self.db_config)

    async def list(self, mt: Any, limit=True) -> Optional[List[Any]]:
        params = self.get_where_params(mt)

        if limit and len(params) < 1:
            return []

        order_by_str = self.get_order_by_str(mt)

        sql = f'''
        select * from {self.table_name} where {self.parse_where_sql(params) or ' 1=1'} {(' order by ' + order_by_str) if order_by_str else ''};
        '''

        return await self._db.query(sql)

    async def insert(self, mt: Any):
        return await self.insert_b([mt])

    async def insert_b(self, mts: List[Any]):
        sql = ''

        for mt in mts:
            ks = ', '.join([field_name for field_name in self.field_names if hasattr(mt, field_name)])
            vs = ', '.join([normallize_insert_value(mt, field_name, getattr(mt, field_name)) for field_name in self.field_names if hasattr(mt, field_name)])

            sql += f'''
            insert into {self.table_name} ({ks}) values ({vs});
            '''

        return await self._db.dml(sql)

    async def update(self, mt: Any, where_mt: Any):
        return await self.update_b([(mt, where_mt)])

    async def update_b(self, args: List[Tuple[Any, Any]]): # mt: Any, where_mt: Any
        sql = ''

        for mt, where_mt in args:
            ks = ', '.join([f'{field_name} = {normallize_insert_value(mt, field_name, getattr(mt, field_name))}' for field_name in self.field_names if hasattr(mt, field_name)])

            params = self.get_where_params(where_mt)

            where_str = '1 = 1'

            if len(params) > 0:
                where_str = self.parse_where_sql(params)

            sql += f'''
            update {self.table_name}
            set {ks}
            where {where_str};
            '''

        return await self._db.dml(sql)

    async def delete(self, mt: Any):
        params = self.get_where_params(mt)

        if len(params) < 1:
            return

        sql = f'''
        delete from {self.table_name} where {self.parse_where_sql(params)};
        '''

        return await self._db.dml(sql)


'''
Author: liubei
Date: 2021-04-09 09:15:30
LastEditTime: 2021-07-02 16:01:18
Description: 
'''
import pprint
import glob
import os
import importlib
import re
import click
from tornado.ioloop import IOLoop
from tornado.web import Application, RequestHandler

from .base.utils import iswin
from config import configs

class TestHandler(RequestHandler):
    def get(self):
        return self.write('test')

def create_app(routes = []):
    pprint.pprint(routes)
    return Application([('/test', TestHandler), *routes], debug=True)

def parse_handler_path(handler_path):
    # '.\\app1\\test_handler.py'
    restful = True if '_restful_handler.py' in handler_path else False
    # 去除restful修饰
    clsname = os.path.basename(handler_path.replace('_restful_handler.py', '_handler.py')).replace('.py', '')
    # TestHandler
    clsname = ''.join(map(lambda x:x.capitalize(), clsname.split("_")))
    # '/app1/test'
    path = handler_path.replace('_restful_handler.py', '') \
            .replace('_handler.py', '') \
            .replace('.\\', '/') \
            .replace('\\', '/')
    # 'app1.test_handler'
    module_path = handler_path.replace('.py', '').replace('.\\', '').replace('\\', '.')

    return clsname, path, module_path, restful

def parse_handler_path_linux(handler_path):
    # './app1/test_handler.py'
    restful = True if '_restful_handler.py' in handler_path else False
    # 去除restful修饰
    clsname = os.path.basename(handler_path.replace('_restful_handler.py', '_handler.py')).replace('.py', '')
    # TestHandler
    clsname = ''.join(map(lambda x:x.capitalize(), clsname.split("_")))
    # '/app1/test'
    path = handler_path.replace('_restful_handler.py', '') \
            .replace('_handler.py', '') \
            .replace('./', '/')
    # 'app1.test_handler'
    module_path = handler_path.replace('.py', '').replace('./', '').replace('/', '.')

    return clsname, path, module_path, restful

def build_routes(apps):
    # [.\\app1, ...]
    sub_dirs = glob.glob('./*')
    handlers = []

    for d in sub_dirs:
        if os.path.isdir(d) and (d.replace('./', '') in apps or d.replace('.\\', '') in apps):
            # ['.\\app1\\test_handler.py', ...]
            handlers.extend(find_handlers(d))

    routes = []

    for h in handlers:
        clsname, path, module_path, restful = parse_handler_path(h) if iswin() else parse_handler_path_linux(h)
        m = importlib.import_module(module_path)

        if hasattr(m, clsname):
            routes.append((path, getattr(m, clsname)))

            if restful:
                routes.append((f'{path}/(.*)', getattr(m, clsname)))

    return routes

def find_handlers(dirname):
    handlers = glob.glob(f'{dirname}/*_handler.py')
    # print(handlers)

    return handlers

@click.group()
def cli1():
    '''
    app相关命令
    '''

@cli1.command()
@click.option('--port', default=9001, type=int, help='默认端口8001')
@click.option('--app', type=str, help='启动的app')
def startapp(port, app):
    if not app or app not in configs['apps']:
        print(f'app: {app} 在config中不存在！')
        exit(1)

    print(f'startApp at {port}...')
    app = create_app(build_routes(configs['apps'][app]))
    app.listen(port)
    IOLoop.current().start()


if __name__ == '__main__':
    print('hi!')
    cli1()



import platform

def dict2class(classT, d):
    '''
    字典类型转换为类对象
    '''
    t = classT()
    t.__dict__.update(d)
    return t

def class2dict(clsi):
    return clsi.__dict__.copy()

def get_module_data_error(module_data, default_msg=None):
    default_msg = default_msg or '获取内部模块数据失败'

    if not module_data:
        return default_msg

    if not module_data.get('status'):
        return module_data.get('msg', default_msg)

def get_mdoule_data(module_data):
    if module_data:
        return module_data.get('data')


def iswin():
    sys = platform.system()
    return sys == 'Windows'

def islinux():
    sys = platform.system()
    return sys == 'Linux'


'''
Author: liubei
Date: 2021-07-02 15:57:21
LastEditTime: 2021-07-03 16:39:39
Description: 
'''
import setuptools

with open("README.md", "r", encoding='utf8') as fh:
    long_description = fh.read()

setuptools.setup(
    name="anbefm",
    version="0.0.4",
    author="liubei",
    author_email="1021086660@qq.com",
    description="一个后端框架",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitee.com/liubei90/basic.git",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    setup_requires=['setuptools', 'wheel', 'twine', 'pipreqs'],
    install_requires=['aiomysql==0.0.21', 'tornado==6.1', 'click==7.1.2', 'PyMySQL==0.9.3'],
)




class CommonException(Exception):
    def __init__(self, val: str):
        self._val = val

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return str(self._val)


class NoModuleException(CommonException):
    def __init__(self, val: str):
        super().__init__()


class NoDataSourceException(CommonException):
    def __init__(self, val: str):
        super().__init__()


class UnavailableDataSourceTypeException(CommonException):
    def __init__(self, val: str):
        super().__init__()



'''
Author: liubei
Date: 2021-04-08 17:21:53
LastEditTime: 2021-04-09 11:48:44
Description: 
'''
import asyncio
import json
from tornado.web import RequestHandler


class PayloadParams():
    ...


class HandlePayload(RequestHandler):
    payload: dict = {}

    async def prepare(self):
        fut = super().prepare()

        if asyncio.coroutines.iscoroutine(fut):
            await fut

        if 'application/json' in self.request.headers.get('Content-type', '') and self.request.body:
            res = '{}'
            print(self.request.body)
            try:
                res = self.request.body.decode('utf-8')
            except:
                res = self.request.body.decode('gbk')

            print(res)
            self.payload = json.loads(res) or {}
